import re
import sqlite3
from collections import namedtuple
from pathlib import Path
from typing import Union, List

DB_PATH = Path(__file__).parent / 'zengin.db'

CONN = sqlite3.connect(DB_PATH.absolute())
DB = CONN.cursor()

kana_table = str.maketrans(
    'ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬねのはばぱひびぴふぶぷへべぺほぼぽ'
    'まみむめもゃやゅゆょよらりるれろわをんーゎゐゑゕゖゔァィゥェォッャュョヮヵヶヰヱヲ',
    'アアイイウウエエオオカガキギクグケゲコゴサザシジスズセゼソゾタダチヂッツヅテデトドナニヌネノハバパヒビピフブプヘベペホボポ'
    'マミムメモヤヤユユヨヨラリルレロワオンーワイエカケヴアイウエオツヤユヨワカケイエオ')


def full_katakana(name: str) -> bool:
    p = re.compile(r'[\u30A1-\u30FF]+')
    m = p.fullmatch(name)
    return m is not None and m.endpos == len(name)


class Zengin:
    """A class for managing DB connections.
    In a multi-threaded environment you don't want to share the default DB
    connection. This class encapsulates a DB connection and can be used as a
    context manager.
    """

    def __init__(self):
        self._conn = sqlite3.connect(DB_PATH)
        self._db = self._conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._conn.close()

    def get(self, code):
        return get(code, self._db)


class Branch(namedtuple('Branch', ['bank_code', 'branch_code', 'branch_name', 'branch_zen_kana', 'branch_han_kana',
                                   'bank_name', 'bank_full_name', 'bank_zen_kana', 'bank_han_kana'],
                        defaults=['', '', '', '', '', '', '', '', ''])):
    @property
    def full_name(self):
        if self.bank_code == '9900':
            return self.branch_name
        if '営業' in self.branch_name:
            return self.branch_name
        if self.branch_name.endswith('店'):
            return self.branch_name
        if self.branch_name.endswith('出張所'):
            return self.branch_name
        return self.branch_name + '支店'

    def __str__(self):
        return self.__repr__()[:-1] + f", full_name='{self.full_name}')"

    @classmethod
    def search(cls, bank_code, name: str):
        where_stmt = 's.name like ?'
        if full_katakana(name.translate(kana_table)):
            name = name.translate(kana_table)
            where_stmt = 's.zen_kana like ?'

        DB.execute("SELECT s.bank_code, s.branch_code, s.name branch_name, s.zen_kana branch_zen_kana, "
                   "  s.han_kana branch_han_kana, b.name bank_name, b.full_name bank_full_name, "
                   "  b.zen_kana bank_zen_kana, b.han_kana bank_han_kana "
                   "FROM bank b INNER JOIN branch s on b.bank_code = s.bank_code "
                   "WHERE s.bank_code=? and " + where_stmt, (bank_code, name + '%',))
        res = DB.fetchall()
        if len(res):
            return [cls(*r) for r in res]
        else:
            return []

    @classmethod
    def get(cls, bank_code: str, branch_code: Union[int, str]) -> List["Branch"]:
        branch_code = str(int(branch_code)).zfill(3)
        if len(branch_code) != 3:
            raise ValueError

        DB.execute("SELECT s.bank_code, s.branch_code, s.name branch_name, s.zen_kana branch_zen_kana, "
                   "  s.han_kana branch_han_kana, b.name bank_name, b.full_name bank_full_name, "
                   "  b.zen_kana bank_zen_kana, b.han_kana bank_han_kana "
                   "FROM bank b INNER JOIN branch s on b.bank_code = s.bank_code "
                   "WHERE s.bank_code=? and s.branch_code=?", (bank_code, branch_code,))
        res = DB.fetchall()
        if len(res):
            return [cls(*r) for r in res]
        else:
            return []


class Bank(namedtuple('Bank', ['bank_code', 'bank_name', 'bank_full_name', 'bank_zen_kana', 'bank_han_kana'])):
    @classmethod
    def get(cls, bank_code: Union[int, str]):
        code = str(int(bank_code)).zfill(4)
        if len(code) != 4:
            raise ValueError

        DB.execute("select bank_code, name, full_name, zen_kana, han_kana from bank where bank_code = ?", (code,))
        res = DB.fetchone()
        if res:
            return cls(*res)

    @classmethod
    def search(cls, name: str):
        if full_katakana(name.translate(kana_table)):
            DB.execute("SELECT bank_code, name, full_name, zen_kana, han_kana FROM bank WHERE zen_kana LIKE ?",
                       ('%' + name.translate(kana_table) + '%',))
        else:
            DB.execute("SELECT bank_code, name, full_name, zen_kana, han_kana FROM bank WHERE name LIKE ?",
                       ('%' + name + '%',))
        res = DB.fetchall()
        if len(res):
            return [cls(*r) for r in res]
        else:
            return []

    @classmethod
    def major_banks(cls):
        DB.execute("select bank_code, name, full_name, zen_kana, han_kana from bank "
                   "where bank_code in ('0001', '0005', '0009', '0010', '0017')")
        res = DB.fetchall()
        if len(res):
            return [cls(*r) for r in res]
        else:
            return []

    @property
    def branches(self):
        DB.execute("select bank_code, branch_code, name, zen_kana, han_kana from branch "
                   "where bank_code = ? order by branch_code",
                   (self.bank_code,))
        res = DB.fetchall()
        if len(res):
            return [Branch(*r)._replace(**self._asdict()) for r in res]


def get(bank_code: str, branch_code):
    return Branch.get(bank_code, branch_code)



import zengin


if __name__ == '__main__':
    for branch in zengin.Branch.get('0001', '001'):
        print(branch)

    for branch in zengin.Branch.search('0005', 'キチジョウジ'):
        print(branch)

    for bank in zengin.Bank.major_banks():
        print(bank)

    for branch in bank.branches:
        print(branch)

    for a in zengin.Bank.search('ユウチヨ'):
        print(a)



import json
import sqlite3
from csv import DictReader
from pathlib import Path

from mojimoji import han_to_zen

FIELDS = [
    'bank_code',
    'branch_code',
    'han_kana',
    'name',
    'branch_flag',  # 名称区分 １：銀行名称 ２：支店名称
    'sub_branch'  # 出張所フラグ １：重複なし or 母店 ２：出張所
]

DB_PATH = Path(__file__).parent / 'zengin.db'


def build_json(bank_csv):
    data = {}
    with open(bank_csv, 'rt', newline='\r\n') as fd:
        for row in DictReader(fd, FIELDS):
            han_kana = row.pop('han_kana').strip()
            row['zen_kana'] = han_to_zen(han_kana)
            row['han_kana'] = han_kana

            if row.pop('branch_flag') == '1':
                row.pop('branch_code')
                row.pop('sub_branch')
                row['branches'] = {}
                data[row['bank_code']] = row
            else:
                branches = data[row.pop('bank_code')]['branches']
                if row['sub_branch'] == '1':
                    branches[row['branch_code']] = row
                else:
                    branch_code = row['branch_code']
                    if 'sub_branches' not in branches[branch_code]:
                        branches[branch_code]['sub_branches'] = []
                    branches[branch_code]['sub_branches'].append(row)

    # same name list
    # 0401 シティバンク、エヌ・エイ
    # 0403 バンク・オブ・アメリカ・エヌ・エイ
    # 0429 バンクネガラインドネシア
    # 0482 アイエヌジー　バンク　エヌ・ヴィ
    # 0484 ナショナル・オーストラリア・バンク・リミテッド
    # 0617 フィリピン・ナショナル・バンク
    # 1000 信金中央金庫
    # 2213 整理回収機構
    for code in ['0401', '0403', '0429', '0482', '0484', '0617', '1000', '2213']:
        bank = data[code]
        bank['full_name'] = bank['name']

    # specific cases
    for code, _, expand in zip(['2004', '2010', '3000', '3771'], ['商工中金', '全信組連', '農林中金', '秋田たかのす'],
                               ['商工組合中央金庫', '全国信用協同組合連合会', '農林中央金庫', '秋田たかのす農業協同組合']):
        bank = data[code]
        bank['full_name'] = expand

    # abbreviation of financial institutions
    for bank in data.values():
        for abbr, expand in zip(
                ['農協', '漁協', '信漁連', '信組', '信連', '労金', '信金'],
                ['農業協同組合', '漁業協同組合', '信用漁業協同組合連合会', '信用組合', '信用連合会', '労働金庫', '信用金庫']):
            if 'full_name' not in bank:
                if bank['name'].endswith(abbr):
                    bank['full_name'] = bank['name'][:-len(abbr)] + expand

    # banks
    for bank in data.values():
        if 'full_name' not in bank:
            bank['full_name'] = bank['name'] + '銀行'

    # write json file
    with open('zengin/zengindata.json', 'w') as outfile:
        outfile.write(json.dumps(data, ensure_ascii=False, indent=2))
    return data


def write_to_sqlite_db(data, db_file=DB_PATH):
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()
    cursor.execute("DROP TABLE IF EXISTS bank")
    cursor.execute("""
        CREATE TABLE bank(
          bank_code TEXT PRIMARY KEY,
          name TEXT, 
          full_name TEXT, 
          zen_kana TEXT, 
          han_kana TEXT
        )""")
    cursor.execute("DROP TABLE IF EXISTS branch")
    cursor.execute("""
        CREATE TABLE branch(
          bank_code TEXT,
          branch_code TEXT, 
          name TEXT, 
          zen_kana TEXT, 
          han_kana TEXT,
          sub_branch TEXT
        )""")

    cursor.execute("CREATE INDEX ix_code ON branch(bank_code, branch_code)")

    for bank in data.values():
        cursor.execute("INSERT INTO bank(bank_code, name, full_name, zen_kana, han_kana)VALUES(?, ?, ?, ?, ?)",
                       (bank['bank_code'], bank['name'], bank['full_name'], bank['zen_kana'], bank['han_kana']))
        for branch in bank['branches'].values():
            cursor.execute("INSERT INTO BRANCH(bank_code, branch_code, name, zen_kana, han_kana, sub_branch) "
                           "VALUES(?, ?, ?, ?, ?, ?)",
                           (bank['bank_code'], branch['branch_code'], branch['name'], branch['zen_kana'],
                            branch['han_kana'], branch['sub_branch']))
            if 'sub_branches' in branch:
                for sub in branch['sub_branches']:
                    cursor.execute("INSERT INTO BRANCH(bank_code, branch_code, name, zen_kana, han_kana, sub_branch) "
                                   "VALUES(?, ?, ?, ?, ?, ?)",
                                   (bank['bank_code'], sub['branch_code'], sub['name'], sub['zen_kana'],
                                    sub['han_kana'], sub['sub_branch']))
    conn.commit()
    conn.close()


if __name__ == '__main__':
    data = build_json('raw/ginkositen.utf8.csv')
    write_to_sqlite_db(data)
